package de.hdm_stuttgart.mi.sd1.tictactoe;

public class Board {

   public enum Player {

      JUNE("June", 'O'), Bill("Bill", 'X');

      public final String nickname;
      public final char representation;

      Player(final String nickname, final char representation) {
         this.nickname = nickname;
         this.representation = representation;
      }

      public Player getOtherPlayer() {
         switch (this) {
         case JUNE:
            return Bill;
         case Bill:
            return JUNE;
         default:
            return null;
         }
      }

      @Override
      public String toString() {
         return "" + representation;
      }
   }
   
   static private int clearTerminalLineCount = 20;                        // Scroll down this many lines simulating terminal clear.
   static private int fieldSize = 3;
   static private int numberOfFields = fieldSize * fieldSize;             // Each Tic-Tac-Toe board has got 3 x 3 = 9 fields
   private short numberOfMoves = 0;                                       // Game will be over after a maximum of 9 moves

   private Player currentActivePlayer;                                    // Player about to start next move

   final Player[] board = new Player[numberOfFields];                     // Allocating a one- dimensional array, yet empty.
      
   public Board(final Player startingPlayer) {
     currentActivePlayer = startingPlayer; 
      for (int i = 0; i < numberOfFields; i++) {                          // Initialize board with null values.
            board[i] = null;
      }
   }

   /**
    * Test whether one of our two players has won the game.
    * 
    * @return The winner instance if a winner exists, null otherwise
    */
   public Player evaluateWinner() {

     if (null != board[4] && (                                            // Check both diagonals
         (board[4] == board[0] && board[4] == board[8]) ||                //
         (board[4] == board[6] && board[4] == board[2]))) {               //   x x
       return board[4];                                                   //    x
     }                                                                    //   x x
     
     for (int i = 0; i < fieldSize; i++) {                                // Check all three columns
       if (null != board[i] &&                                            //  +-=                                      
           board[i] == board[i + fieldSize] &&
           board[i] == board[i + 2* fieldSize]) {                         //  +-=
         return board[i];                                                 //  +-=
       }
     }
     
     for (int i = 0; i < 2 * fieldSize + 1; i += fieldSize) {             // Check all three rows
       if (null != board[i] &&                                            // +++
           board[i] == board[i + 1] && board[i] == board[i + 2]) {        // ---
         return board[i];                                                 // ===
       }
     }
     return null;
   }

   /**
    * Occupy the next field by player {@link #getCurrentActivePlayer()}.
    * 
    * Fields are being numbered by:
    * 
    *  1|2|3
    *  -+-+-
    *  4|5|6
    *  -+-+-
    *  7|8|9
    * 
    * @param field The field in question
    *  
    * @return An error message indication either an occupied field or an index violation.
    *          null if everything is o.K.
    */
   public String nextMove(final int field) {                          

      if (field < 1 || numberOfFields < field) {
         return "Field index out of range [0 ..." + 
               numberOfFields + "]";
      } else {
        final int fieldIndex = field - 1;                               // User to (computer) nerd point of view
         final Player current = board[fieldIndex];
         if (current == null) {
            board[fieldIndex] = currentActivePlayer;
            currentActivePlayer = currentActivePlayer.getOtherPlayer();
            numberOfMoves++;
            return null;
         } else {
            return "Field already occupied by " + current.nickname;
         }
      }
   }
   
   static public void printNumberingHint() {
      System.out.println("Numbering scheme:\n");
      for (int y = 0; y < 3; y++) {
         for (int x = 0; x < 3; x++) {
            System.out.print((char)('1' + x + y * fieldSize));
            if (x < fieldSize - 1) {
               System.out.print('|');
            }
         }
         if (y < fieldSize - 1) {
            System.out.println("\n-+-+-");
         }
      }
   }
   
   /**
    * Get the player instance which will do the next move. See {@link #nextMove(short)}
    * 
    * @return the currently active player.
    */
   public Player getCurrentActivePlayer() {
      return currentActivePlayer;
   }
   
   /**
    * After nine moves all fields will have been occupied.
    * @return true if all possible moves have been completed, false otherwise.
    */
   public boolean allMovesFinished() {
      return numberOfFields <= numberOfMoves;
   }
   
   /**
    * Publish the current state to standard output using
    * ASCII graphics.
    */
   public void print() {
      for (int clear = 0; clear < clearTerminalLineCount; clear++) {
         System.out.println("\n");
      }
      
      System.out.println("\n\nPlayer " + Player.JUNE.nickname + "("
            + Player.JUNE.representation + ")" + "\nvs. " + Player.Bill.nickname
            + "(" + Player.Bill.representation + ")"  + "          Free fields\n");
      
      final String tableSeparator = "                  ";
      for (int x = 0; x < fieldSize; x++) {
         for (int y = 0; y < fieldSize; y++) {
           final int fieldIndex = y + fieldSize * x;
            System.out.print(null == board[fieldIndex] ? " " : board[fieldIndex]);
            if (y < fieldSize - 1) {
               System.out.print('|'  );
            } 
         }
         System.out.print(tableSeparator);
         
         for (int y = 0; y < fieldSize; y++) {
           final int fieldIndex = y + fieldSize * x;
            System.out.print(null == board[fieldIndex] ? fieldIndex + 1 : // Nerd to user perspective.
              " ");
            if (y < fieldSize - 1) {
               System.out.print('|');
            }
         }

         if (x < fieldSize - 1) {
            System.out.println("\n-+-+-" + tableSeparator + "-+-+-");
         }
      }
   }
}
